package Plugins::MyPicks::Plugin;
#
# MyPicks Copyright (C) Bryan Alton, Triode 2006
#
# Customisable My Picks Browser and Editor
#
# A Slimserver plugin for creating and customising internet radio lists
#
#	Version History
#	0.1		- Initial Release
#   0.2     - Rewrite of UI to use web template code rather than setup interface
#   0.3     - Minor bug fixes, updated cosmetic changes to templates, thanks to input from kdf
#   0.4     - Added ability to import OPML files to current level, export current level to favorites
#           - Error message if no mypicks file found on player, jump to editor on web interface
#           - Added cli support for mypicks list
#   0.5     - ascend levels in edit mode via breadcrumb list
#           - allow import of M3U and PLS playlists containing local files
#           - escape file:// urls correctly
#   0.6     - bugfix to html to allow use with IE7

use strict;

use Slim::Buttons::Common;
use Slim::Utils::Misc;
use Slim::Utils::Strings qw(string);
use Slim::Utils::Cache;
use Slim::Utils::Favorites;

use XML::Simple;
use File::Basename;
use File::Spec::Functions qw(:ALL);
use File::Temp qw(tempfile);
use IO::Socket qw(:crlf);

use vars qw($VERSION);
$VERSION = "0.5";

my $myPicksName = 'mypicks.opml';

# wiki used as temporary hosting service for restructured slim picks!
my $directoryURL = 'http://wiki.slimdevices.com/plugin/attachments/RadioStationOPMLs/directory.opml';

my $cache_time = 30*60; # time to cache directory opml before fetching again - 30 mins so fetch once while editing

my $opml;
my $directoryopml;
my $level = 0;
my $currentLevel;
my @prevLevels;
my $deleted;
my $cli_next;

sub enabled {
    return ( $::VERSION ge '6.5' );
}

sub getDisplayName {
    return 'PLUGIN_MYPICKS';
}

sub addMenu {
    return 'RADIO';
}

sub getFunctions {
    return { 'left' => sub { Slim::Buttons::Common::popModeRight(@_[0]) } };
}

sub initPlugin {
    $::d_plugins && msg("MyPicks Plugin initializing.\n");

    Slim::Buttons::Common::addMode( 'PLUGIN.MyPicks', getFunctions(), \&setMode );

#        |requires Client
#        |  |is a Query
#        |  |  |has Tags
#        |  |  |  |Function to call
#        C  Q  T  F
    Slim::Control::Request::addDispatch(['mypicks', 'items', '_index', '_quantity'],
        [0, 1, 1, \&cliQuery]);
	Slim::Control::Request::addDispatch(['mypicks', 'playlist', '_method' ],
		[1, 1, 1, \&cliQuery]);
	$cli_next=Slim::Control::Request::addDispatch(['radios', '_index', '_quantity' ],
		[0, 1, 1, \&cliRadiosQuery]);

    ($opml, undef) = loadOpmlFile();
}

sub setMode {
    my $client = shift;
    my $method = shift;

    if ( $method eq 'pop' ) {
        Slim::Buttons::Common::popMode($client);
        return;
    }

	my $file = makeFileName($myPicksName);

	if (-r $file) {

		# use INPUT.Choice to display the list of feeds
		my %params = (
			header   => 'PLUGIN_MYPICKS_LOADING_PICKS',
			modeName => 'MyPicks Plugin',
			url      => Slim::Utils::Misc::fileURLFromPath($file),
			title    => $client->string( getDisplayName() ),
		   );
		
		Slim::Buttons::Common::pushMode($client, 'xmlbrowser', \%params);
	
		# we'll handle the push in a callback
		$client->param('handledTransition',1)

	} else {

		$client->lines(\&errorLines);
	}
}

sub errorLines {
	return { 'line' => [string('PLUGIN_MYPICKS'), string('PLUGIN_MYPICKS_NOMYPICKS')] };
}

sub cliQuery {
	my $request = shift;

	$::d_plugins && msg("MyPicks: cliQuery()\n");

	my $file = makeFileName($myPicksName);

	Slim::Buttons::XMLBrowser::cliQuery('mypicks', Slim::Utils::Misc::fileURLFromPath($file), $request);
}

sub cliRadiosQuery {
	my $request = shift;

	$::d_plugins && msg("MyPicks: cliRadiosQuery()\n");

	# what we want the query to report about ourself
	my $data = {
		'cmd' => 'mypicks',                    # cmd label
		'name' => Slim::Utils::Strings::string(getDisplayName()),  # nice name
		'type' => 'xmlbrowser',              # type
	};
	
	# let our super duper function do all the hard work
	Slim::Control::Queries::dynamicAutoQuery($request, 'radios', $cli_next, $data);
}

sub webPages {
	my $title = 'PLUGIN_MYPICKS';
	my $editTitle = 'PLUGIN_MYPICKS_EDITOR';

	if (grep {$_ eq 'MyPicks::Plugin'} Slim::Utils::Prefs::getArray('disabledplugins')) {
		Slim::Web::Pages->addPageLinks('radio', { $title => undef });
		Slim::Web::Pages->addPageLinks('plugins', { $editTitle => undef });
	} else {
		Slim::Web::Pages->addPageLinks('radio', { $title => 'plugins/MyPicks/index.html' });
		Slim::Web::Pages->addPageLinks('plugins', { $editTitle => 'plugins/MyPicks/edit.html' });
	}

	my %pages = ( 

	    'edit.html' => \&editHandler,

		'index.html' => sub {

			my $file = makeFileName($myPicksName);

			if (-r $file) {
 			
				Slim::Web::XMLBrowser->handleWebIndex( {
					feed   => Slim::Utils::Misc::fileURLFromPath($file),
					title  => $title,
					args   => \@_
				   } );

			} else {
				
				return editHandler(@_);
								
			}
				
		},
	);
	
	return \%pages;
}

sub editHandler {
	my ($client, $params) = @_;

	my $edit; # index of entry to edit if set
	my $errorMsg; # error message to display at top of page

	# Debug:
	#for my $key (keys %$params) {
	#	print "Key: $key, Val: ".$params->{$key}."\n";
	#}

	# action any params set
	$opml->{'head'}->{'title'} = $params->{'title'} if $params->{'title'};

	if ($params->{'load'} || $params->{'loadmypicks'}) {
		my $filename = $params->{'load'} ? $params->{'filename'} : $myPicksName;
		($opml, $errorMsg) = loadOpmlFile( makeFileName($filename) );
		$level = 0;
		$currentLevel = undef;
		@prevLevels = ();
		$deleted = undef;
	}

	if ($params->{'save'} || $params->{'savemypicks'}) {
		my $filename = $params->{'save'} ? $params->{'filename'} : $myPicksName;
		$errorMsg = writeOpmlFile( makeFileName($filename), $opml);
		$deleted = undef;
	}

	if ($params->{'loadopml'}) { $params->{'loaddialog'}   = 1; }
	if ($params->{'saveopml'}) { $params->{'savedialog'}   = 1; }
	if ($params->{'import'})   { $params->{'importdialog'} = 1; }

	if ($params->{'export'}) {
		$params->{'exportdialog'} = 1;
		if (exists &Plugins::WikiRadio::wikiRadioTitles) {
			my $title = $opml->{'head'}->{'title'};
			my $wikiTitles = Plugins::WikiRadio::wikiRadioTitles();
			my $titleClash;
			foreach my $entry (@$wikiTitles) {
				$titleClash = 1 if ($entry eq $title);
			}
			if ($titleClash) {
				$params->{'wikiexists'} = 1;
				Plugins::WikiRadio::wikiRadioFlushCache() if exists &Plugins::WikiRadio::wikiRadioFlushCache;
			} else {
				$params->{'wikiexport'} = 1;
			}
		}
	}

	$currentLevel ||= $opml->{'body'}[0]->{'outline'} ||= [];

	if (my $action = $params->{'action'}) {
		if ($action eq 'descend') {
			$prevLevels[ $level++ ] = {
				'ref'   => $currentLevel,
				'title' => @$currentLevel[$params->{'entry'}]->{'text'},
			};
			$currentLevel = @$currentLevel[$params->{'entry'}]->{'outline'};
		}

		if ($action eq 'ascend') {
			my $pop = defined ($params->{'levels'}) ? $params->{'levels'} : 1;
			while ($pop) {
				$currentLevel = $prevLevels[ --$level ]->{'ref'} if $level > 0;
				--$pop;
			}
		}

		if ($action eq 'edit') {
			$edit = $params->{'entry'};
		}

		if ($action eq 'edittitle') {
			$params->{'edittitle'} = 1;
		}

		if ($action eq 'delete') {
			$deleted = splice @$currentLevel, $params->{'entry'}, 1;
		}

		if ($action eq 'forgetdelete') {
			$deleted = undef;
		}

		if ($action eq 'insert' && $deleted) {
			push @$currentLevel, $deleted;
			$deleted = undef;
		}

		if ($action eq 'movedown') {
			my $entry = splice @$currentLevel, $params->{'entry'}, 1;
			splice @$currentLevel, $params->{'entry'} + 1, 0, $entry;
		}

		if ($action eq 'moveup' && $params->{'entry'} > 0) {
			my $entry = splice @$currentLevel, $params->{'entry'}, 1;
			splice @$currentLevel, $params->{'entry'} - 1, 0, $entry;
		}

		if ($action eq 'newentry') {
			my $cache = Slim::Utils::Cache->new();
			$directoryopml = $cache->get( 'mypicks_directory_opml' ) || loadPickDirectory($directoryURL);
			my ($dirmenus, $dirstations) = directoryTitles($directoryopml);

			$params->{'newentry'} = {
				'dirmenus'    => $dirmenus,
				'dirstations' => $dirstations,
			};

			my $wikiRadioTitles = (exists &Plugins::WikiRadio::wikiRadioTitles) ?
				Plugins::WikiRadio::wikiRadioTitles() : undef;
			if ($wikiRadioTitles) {
				splice @$wikiRadioTitles, 0, 0, string('PLUGIN_MYPICKS_SELECT_MENU');
				$params->{'newentry'}->{'wikiradio'} = $wikiRadioTitles;
			}
		}
		
		if ($action =~ /play|add/ && $client) {
			my $stream = @$currentLevel[$params->{'entry'}]->{'URL'};
			my $title  = @$currentLevel[$params->{'entry'}]->{'text'};
			Slim::Music::Info::setTitle($stream, $title);
			$client->execute(['playlist', $action, $stream]);
		}
	}

	if ($params->{'editset'} && defined $params->{'entry'}) {
		my $entry = @$currentLevel[$params->{'entry'}];
		$entry->{'text'} = $params->{'entrytitle'};
		$entry->{'URL'} = $params->{'entryurl'} if defined($params->{'entryurl'});
	}

	if ($params->{'newmenu'}) {
		push @$currentLevel, {
			'text'   => $params->{'menutitle'},
			'outline'=> [],
		};
	}

	if ($params->{'newstream'}) {
		push @$currentLevel,{
			'text' => $params->{'streamtitle'},
			'URL'  => $params->{'streamurl'},
			'type' => 'audio',
		};
	}
		
	if ($params->{'newopmlmenu'}) {
		push @$currentLevel, {
			'text' => $params->{'opmlmenutitle'},
			'URL'  => $params->{'opmlmenuurl'},
		};
	}

	if ($params->{'newdirstream'}) {
		my $new = directoryTitleToEntry($directoryopml, $params->{'dirstream'}, 'station');
		push @$currentLevel, $new if $new;
	}

	if ($params->{'newdirmenu'}) {
		my $new = directoryTitleToEntry($directoryopml, $params->{'dirmenu'}, 'menu');
		push @$currentLevel, $new if $new;
	}

	if ($params->{'newwikimenu'}) {
		my $newurl = exists &Plugins::WikiRadio::wikiRadioTitleToUrl ? 
			Plugins::WikiRadio::wikiRadioTitleToUrl($params->{'wikimenu'}) : undef;

		push @$currentLevel, {
			'text' => $params->{'wikimenu'},
			'URL'  => $newurl,
		} if $newurl;
	}

	if ($params->{'importfav'}) {
		# import favorites
		my $favs = Slim::Utils::Favorites->new($client);
		my @titles = $favs->titles();
		my @urls = $favs->urls();
		my $count = 0;

		foreach my $title (@titles) {
			push @$currentLevel, {
				'text' => $title || string('EMPTY'),
				'URL'  => $urls[$count++],
				'type' => 'audio',
			};
		}
	}

	if ($params->{'importplaylist'}) {
		# import playlist
		my $file = makeFileName($params->{'filename'});
		my ($titles, $urls, $opmltree);
		my $count = 0;

		$::d_plugins && msg("MyPicks: Processing playlistfile '$file'\n");

		if (-r $file) {

			if (Slim::Music::Info::isM3U($file)) {
  				$::d_plugins && msg("MyPicks: import Playlist -  Found m3u file '$file'\n");
				my $fh = new FileHandle;
				$fh->open("<$file");
		 		($urls, $titles) = readM3Uplaylist($fh);
			    $fh->close();

		    } elsif (Slim::Music::Info::isPLS($file)) {
  				$::d_plugins && msg("MyPicks: import Playlist -  Found pls file '$file'\n");
				my $fh = new FileHandle;
				$fh->open("<$file");
				($urls, $titles) = readPLSplaylist($fh);
			    $fh->close();

		    } elsif ($file =~ /\.opml$/) {
  				$::d_plugins && msg("MyPicks: import Playlist -  Found opml file '$file'\n");
				($opmltree, $errorMsg) = loadOpmlFile($file);
			}

    	}
		
		foreach (@$titles) {
			$::d_plugins && msg("MyPicks: import Playlist -  adding entry $_  ($urls->[$count])\n");
			push @$currentLevel, {
				'text' => defined ($_ ) ?  $_ : string('PLUGIN_MYPICKS_NOTITLE')."[$count]",
				'URL'  => $urls->[$count++],
				'type' => 'audio',
			};
		}

		if ($opmltree && $opmltree->{'body'} && ref($opmltree->{'body'}) eq 'ARRAY' && $opmltree->{'body'}[0]->{'outline'}) {
			push @$currentLevel, {
				'text'    => $opmltree->{'head'}->{'title'},
				'outline' => $opmltree->{'body'}[0]->{'outline'},
			};

		} elsif ($count == 0) { 
			$errorMsg ||= string('PLUGIN_MYPICKS_IMPORTERROR') . " - $file";
		}
    }

	if ($params->{'exportfav'}) {
		# export favorites
		my (@titles, @urls);
		
		foreach my $entry (@$currentLevel) {
			if ($entry->{'type'} eq 'audio') {
				push @titles, $entry->{'text'};
				push @urls, $entry->{'URL'};
			}
		}

		# NB writes direct to prefs as no favorite api for this!
		Slim::Utils::Prefs::setArray( 'favorite_titles', \@titles );
		Slim::Utils::Prefs::setArray( 'favorite_urls', \@urls );

		$errorMsg = string('PLUGIN_MYPICKS_FAVEXPORTED');
	}

	if ($params->{'exporttowiki'}) {
		$errorMsg = wikiRadioUpload($opml, $opml->{'head'}->{'title'});
		$errorMsg ||= string('PLUGIN_MYPICKS_FILEUPLOADED');
	}

	# set params for page build
	$params->{'title'} = $opml->{'head'}->{'title'};
	$params->{'previous'} = ($level > 0);
	$params->{'deleted'} = defined($deleted) ? $deleted->{'text'} : undef;
	$params->{'errormsg'} = $errorMsg;

	my @entries;
	my $index = 0;
	
	foreach my $opmlEntry (@$currentLevel) {
		push @entries, {
			'title'   => $opmlEntry->{'text'} || '',
			'url'     => $opmlEntry->{'URL'} || '',
			'audio'   => (defined $opmlEntry->{'type'} && $opmlEntry->{'type'} eq 'audio'),
			'outline' => $opmlEntry->{'outline'},
			'edit'    => (defined $edit && $edit == $index),
			'index'   => $index++,
		};
	}

	$params->{'entries'} = \@entries;

	push @{$params->{'pwd_list'}}, {
		'title' => $opml->{'head'}->{'title'} || string('PLUGIN_MYPICKS_EDITOR'),
		'href'  => 'href="edit.html?action=ascend&levels=' . $level . '"',
	};
	for (my $i = 1; $i <= $level; $i++) {
		push @{$params->{'pwd_list'}}, {
			'title' => $prevLevels[ $i - 1 ]->{'title'},
			'href'  => 'href="edit.html?action=ascend&levels=' . ($level - $i) . '"',
		};
	}

	return Slim::Web::HTTP::filltemplatefile('plugins/MyPicks/edit.html', $params);
}

sub makeFileName {
	my $name = shift;

	if ($name =~ /^file\:\/\/(.*)/) {
		$name = $1;
	} elsif (dirname($name) eq '.') {
		$name = catdir(Slim::Utils::Prefs::get("playlistdir"), $name);
	}

	return $name;
}

sub loadOpmlFile {
    my $filename = shift;
    my $fileopml;
	my $errorMsg;

    if (defined $filename) {

		$fileopml = eval { XMLin( $filename, forcearray => [ 'outline', 'body' ], SuppressEmpty => undef ) };

		if (defined($fileopml)) {
			$::d_plugins && msg("Success: MyPicks XMLin'ed file $filename\n");
		} else {
			$::d_plugins && msg("Failure: MyPicks couldn't XMLin file $filename\n$!\n");
			$errorMsg = string('PLUGIN_MYPICKS_LOADERROR')." - $filename";
		}
    }
		
	$fileopml ||= {
		'head' => {
			'title'          => string('PLUGIN_MYPICKS_NOTITLE'),
			'expansionState' => {},
		},
		'version' => '1.0',
		'body'    => [
			{
				'outline' => [],
			}
		],
	};
	
    return $fileopml, $errorMsg;
}

sub writeOpmlFile {
	my $filename = shift;
	my $opml = shift;

	# ensure server XML cache for this filename is removed
	my $cache = Slim::Utils::Cache->new();
	$cache->remove( Slim::Utils::Misc::fileURLFromPath($filename) . '_parsedXML' );

    my $dir = dirname($filename);

	if (-w $dir) {
	
		my ($tmp, $tmpfilename) = tempfile(TEMPLATE => 'MyPicksTempXXXXX', DIR => $dir, SUFFIX => '.opml', UNLNK => 0, OPEN => 0);

		my $ret = eval { XMLout( $opml, OutputFile => $tmpfilename, XMLDecl => '<?xml version="1.0" encoding="UTF-8"?>', 
								 KeyAttr => ['head', 'body'], SuppressEmpty => undef, Rootname =>"opml", NoSort => 0 ); };

		if (defined($ret)){

			$::d_plugins && msg( "MyPicks: OPML saved to file tempfile: $tmpfilename\n" );

			if ( -e $filename && !rename($filename, $filename . ".backup") ) {
				$::d_plugins && msg("MyPicks: Failed to rename old $filename to backup\n");
			}

			if ( rename($tmpfilename, $filename)) {
				$::d_plugins && msg("MyPicks: Renamed $tmpfilename to $filename\n");
				return undef;

			} else {
				$::d_plugins && msg("MyPicks: Failed for rename $tmpfilename to $filename\n");
			}
		}
	}

	return string('PLUGIN_MYPICKS_SAVEERROR') . " - $filename";
}

sub loadPickDirectory {
	my $url = shift;
	my $opml;
        
	my $http = Slim::Player::Protocols::HTTP->new( { 'url' => $url, 'create' => 0, 'timeout' => 10 } );

	if ( defined $http ) {

		my $content = $http->content;
		$http->close;

		if ( defined $content ) {

			$opml = eval { XMLin( \$content, forcearray => [ 'outline', 'body' ], SuppressEmpty => undef ) };

			if ($@) {

				$::d_plugins && msg("MyPicks: Failed to parse directory OPML <$url> because: $@\n");

			} else {

				my $cache = Slim::Utils::Cache->new();
				$cache->set( 'mypicks_directory_opml', $opml, $cache_time );

			}
		}
	}
		
	if ( !defined $opml ) {

		$::d_plugins && msg("MyPicks: Failed to load OPML direcrory URL <$url> \n");

		$opml = {
            'head' => {
                'title'          => "Unable to load $url",
                'expansionState' => {},
            },
            'version' => '1.0',
            'body'    => [
				 {
					 'outline' => [
					     {
							 'text'    => 'Menus',
							 'outline' => [],
						 },
						 {
							 'text'    => 'Radio Stations',
							 'outline' => [],
						 },
                    ],
                }
            ],
        };
	}

    return $opml;
}

sub directoryTitles {
    my $diropml = shift;

    my ( @menus, @stations );

    push @menus, string('PLUGIN_MYPICKS_SELECT_MENU');
    push @stations, string('PLUGIN_MYPICKS_SELECT_STATION');

    return ( \@menus, \@stations ) unless defined($diropml);

    # Process menu first
    my $ptr = $diropml->{'body'}[0]->{'outline'}[0]->{'outline'};
    foreach my $item ( @{$ptr} ) {
        push @menus, $item->{'text'};
    }

    # Process stations next
    $ptr = $diropml->{'body'}[0]->{'outline'}[1]->{'outline'};
    foreach my $item ( @{$ptr} ) {
        push @stations, $item->{'text'};
    }

    return ( \@menus, \@stations );
}

sub directoryTitleToEntry {
	my $diropml = shift;
	my $title = shift;
	my $level = shift eq 'menu' ? 0 : 1;

    my $ptr = $diropml->{'body'}[0]->{'outline'}[$level]->{'outline'};
    foreach my $item ( @{$ptr} ) {
        if ($item->{'text'} eq $title) {
			return $item;
		}
    }
	return undef;
}


# The following playlist parsers are modified versions of the parsers in the main slimserver code
# They are modified to return title and url without touching the database...

sub readM3Uplaylist {
	my $m3u    = shift;

	my @items  = ();
	my @urls   = ();
	my @titles = (); 
	my ($secs, $artist, $album, $title);
	my $foundBOM = 0;

	$::d_plugins && msg("MyPicks: parsing M3U: $m3u\n");

	while (my $entry = <$m3u>) {

		chomp($entry);

		# strip carriage return from dos playlists
		$entry =~ s/\cM//g;  

		# strip whitespace from beginning and end
		$entry =~ s/^\s*//; 
		$entry =~ s/\s*$//; 
		
		# decode any HTML entities in-place
		HTML::Entities::decode_entities($entry);

		if (!$foundBOM) {

			$entry = Slim::Utils::Unicode::stripBOM($entry);
			$foundBOM = 1;
		}

		# Guess the encoding of each line in the file. Bug 1876
		# includes a playlist that has latin1 titles, and utf8 paths.
		my $enc = Slim::Utils::Unicode::encodingFromString($entry);

		$entry = Slim::Utils::Unicode::utf8decode_guess($entry, $enc);

		$::d_plugins && msg("MyPicks M3U Parse:  entry from file: $entry\n");

		if ($entry =~ /^#EXTINF\:(.*?),<(.*?)> - <(.*?)> - <(.*?)>/) {
			$secs   = $1;
			$artist = $2;
			$album  = $3;
			$title  = $4;
			$::d_plugins && msg("MyPicks M3U Parse:  found secs: $secs, title: $title, artist: $artist, album: $album\n");

		} elsif ($entry =~ /^#EXTINF:(.*?),(.*)$/) {
			$secs  = $1;
			$title = $2;	
			$::d_plugins && msg("MyPicks M3U Parse:  found secs: $secs, title: $title\n");
		}

		next if $entry =~ /^#/;
		next if $entry =~ /#CURTRACK/;
		next if $entry eq "";


		$entry =~ s|$LF||g;

		$entry = Slim::Utils::Unicode::utf8encode_locale($entry);

		$::d_plugins && msg("MyPicks M3U Parse:    entry: $entry\n");
 
	    if (Slim::Music::Info::isURL($entry)) {
			$::d_plugins && msg("MyPicks M3U Parse: Valid URL Adding entry: $entry\n");
			push @urls, $entry;
			push @titles, $title;

		} elsif (-r $entry) {
			$::d_plugins && msg("MyPicks M3U Parse: Valid File Adding entry: $entry\n");
			push @urls, Slim::Utils::Misc::fileURLFromPath($entry);
			push @titles, $title;
		}
	}

	$::d_plugins && msg("MyPicks M3U Parse: parsed " . scalar(@items) . " items in m3u playlist\n");
	return (\@urls, \@titles);
}

sub readPLSplaylist {
	my $pls    = shift;

	my @plsurls   = ();
	my @plstitles = ();
	my @urls   = ();
	my @titles = ();

	my @items  = ();
	my $foundBOM = 0;
	
	$::d_plugins && msg("MyPicks: parsing PLS: $pls \n");
	
	while (my $line = <$pls>) {
		chomp($line);
		$::d_plugins && msg("MyPicks: parsing PLS: Parsing line: $line\n");

		# strip carriage return from dos playlists
		$line =~ s/\cM//g;

		# strip whitespace from end
		$line =~ s/\s*$//;

		if (!$foundBOM) {
			$line = Slim::Utils::Unicode::stripBOM($line);
			$foundBOM = 1;
		}

		# Guess the encoding of each line in the file. Bug 1876
		# includes a playlist that has latin1 titles, and utf8 paths.
		my $enc = Slim::Utils::Unicode::encodingFromString($line);

		$line = Slim::Utils::Unicode::utf8decode_guess($line, $enc);
	
		if ($line =~ m|File(\d+)=(.*)|i) {
			$plsurls[$1] = Slim::Utils::Unicode::utf8encode_locale($2);
			next;
		}

		if ($line =~ m|Title(\d+)=(.*)|i) {
			$plstitles[$1] = $2;
			next;
		}	
	}

	for (my $i = 1; $i <= $#plsurls; $i++) {
		if (Slim::Music::Info::isURL($plsurls[$i])) {
			$::d_plugins && msg("MyPicks: parsing PLS: found URL entry $i " .  $plstitles[$i] . "(".  $plsurls[$i] .")\n") ;
			push @urls, $plsurls[$i];
			push @titles, $plstitles[$i];
		} elsif (-r $plsurls[$i]) {
			$::d_plugins && msg("MyPicks: parsing PLS: found File entry $i " .  $plstitles[$i] . "(".  $plsurls[$i] .")\n") ;
			push @urls, Slim::Utils::Misc::fileURLFromPath($plsurls[$i]);
			push @titles, $plstitles[$i];
		}

	}

	$::d_plugins && msg("MyPicks PLS Parse: parsed " . scalar(@items) . " items in pls playlist\n");
	
	return (\@urls, \@titles);
}

sub wikiRadioUpload {
	my $opml = shift;
	my $title = shift;

	return string('PLUGIN_MYPICKS_NOWIKIRADIO') unless exists &Plugins::WikiRadio::wikiRadioUpload;

	my $tmpfilename = catdir(Slim::Utils::Prefs::get('cachedir'), 'wikiradiotmp.opml');
	my $title = $title . '.opml';

	$::d_plugins && msg("MyPicks: Uploading to wikiradio\n");
	
	XMLout( $opml, OutputFile => $tmpfilename, XMLDecl => '<?xml version="1.0" encoding="UTF-8"?>', 
		KeyAttr => ['head', 'body'], SuppressEmpty => undef, Rootname =>"opml", NoSort => 0 );

	my $errorMsg = Plugins::WikiRadio::wikiRadioUpload($tmpfilename, $title);

	$::d_plugins && msg("MyPicks: Upload Error - $errorMsg\n") if $errorMsg;

	Plugins::WikiRadio::wikiRadioFlushCache() if exists &Plugins::WikiRadio::wikiRadioFlushCache;

	unlink($tmpfilename);

	return $errorMsg;
}
	
sub strings {
    return '
PLUGIN_MYPICKS
	DE	Meine Internet-Radiostationen
	EN	My Picks

PLUGIN_MYPICKS_EDITOR
	DE	Meine Internet-Radiostationen bearbeiten
	EN	My Picks OPML Editor

PLUGIN_MYPICKS_LOADING_PICKS
	DE	Meine Internet-Radiostationen laden
	EN	Loading My Picks

PLUGIN_MYPICKS_CANCEL
	DE	Abbrechen
	EN	Cancel

PLUGIN_MYPICKS_SET
	DE	Übernehmen
	EN	Set

PLUGIN_MYPICKS_LOAD
	DE	Laden
	EN	Load

PLUGIN_MYPICKS_SAVE
	DE	Speichern
	EN	Save

PLUGIN_MYPICKS_CHANGE
	DE	Ändern
	EN	Change

PLUGIN_MYPICKS_ADD_DESC
	DE	Eines der folgenden Elemente hinzufügen
	EN	Add one of the following

PLUGIN_MYPICKS_ADDMENU
	DE	Neues Menü
	EN	New menu

PLUGIN_MYPICKS_ADDSTREAM
	DE	Audio Stream
	EN	Audio stream

PLUGIN_MYPICKS_ADDOPMLMENU
	DE	OPML Menü auf anderem Server
	EN	Remote OPML menu

PLUGIN_MYPICKS_ADDSLIMPICKS
	DE	Eintrag aus der Slim Devices Auswahl
	EN	Entry from Slim Picks

PLUGIN_MYPICKS_ADDWIKIRADIO
	DE	Eintrag aus dem Wiki Radio
	EN	Entry from Wiki Radio

PLUGIN_MYPICKS_LOADMYPICKS
	DE	Meine Stationen laden
	EN	Load My Picks

PLUGIN_MYPICKS_SAVEMYPICKS
	DE	Meine Stationen speichern
	EN	Save My Picks

PLUGIN_MYPICKS_LOADFILE
	DE	Datei laden
	EN	Load File

PLUGIN_MYPICKS_SAVEFILE
	DE	Datei speichern
	EN	Save File

PLUGIN_MYPICKS_IMPORT
	DE	Importieren
	EN	Import

PLUGIN_MYPICKS_EXPORT
	DE	Exportieren
	EN	Export

PLUGIN_MYPICKS_IMPORTFAV
	DE	Favoriten importieren
	EN	Import Favorites

PLUGIN_MYPICKS_EXPORTFAV
	DE	Als Favoriten exportieren
	EN	Export Favorites

PLUGIN_MYPICKS_FAVEXPORTED
	DE	Favoriten exportiert
	EN	Favorites Exported

PLUGIN_MYPICKS_IMPORT_DESC
	DE	Wiedergabeliste (M3U, PLS oder OPML) oder Favoriten zum aktuellen Menü hinzufügen
	EN	Import a playlist (M3U, PLS or OPML) or your Favorites to the current menu level

PLUGIN_MYPICKS_EXPORT_FAVDESC
	DE	Sie können das aktuelle Menüs als Favoriten exportieren. Dies überschreibt die aktuellen Favoriten.
	EN	You may export the current level to your Favorites. This will overwrite your current Favorites.

PLUGIN_MYPICKS_EXPORT_WIKIDESC
	DE	Sie können ihr ganzes Menü aufs <a href="http://wiki.slimdevices.com/index.cgi?WikiRadio" target="WikiRadio"><u>WikiRadio</u></a> hochladen, um ihre Auswahl so allen SlimServer zugänglich zu machen.
	EN	You may upload all levels to share your stations with other Slim users via <a href="http://wiki.slimdevices.com/index.cgi?WikiRadio" target="WikiRadio"><u>WikiRadio</u></a>.

PLUGIN_MYPICKS_PLAYLISTFILE
	DE	Wiedergabeliste
	EN	Playlist File

PLUGIN_MYPICKS_PREVLEVEL
	DE	Vorheriges Niveau
	EN	Previous Level

PLUGIN_MYPICKS_INSERTDEL
	DE	Gelöschtes Element einfügen
	EN	Insert Deleted Entry

PLUGIN_MYPICKS_NEWENTRY
	DE	Neues Element
	EN	New Entry

PLUGIN_MYPICKS_NOTITLE
	DE	Kein Titel
	EN	No Title

PLUGIN_MYPICKS_SELECT_MENU
	DE	-- Wähle Menü --
	EN	-- Select Menu --

PLUGIN_MYPICKS_SELECT_STATION
	DE	-- Wähle Radiostation --
	EN	-- Select Station --

PLUGIN_MYPICKS_LOADERROR
	DE	Fehler beim Laden der Datei
	EN	Error Loading File

PLUGIN_MYPICKS_SAVEERROR
	DE	Fehler beim Speichern der Datei
	EN	Error Saving File

PLUGIN_MYPICKS_IMPORTERROR
	DE	Fehler beim Importieren der Datei
	EN	Error Importing File

PLUGIN_MYPICKS_UPLOAD
	DE	Zum WikiRadio hochladen.
	EN	Upload To WikiRadio

PLUGIN_MYPICKS_FILEUPLOADED
	DE	Datei übertragen
	EN	File Uploaded

PLUGIN_MYPICKS_NOWIKIRADIO
	DE	Das WikiRadio Plugin ist nicht geladen
	EN	The WikiRadio Plugin is not loaded

PLUGIN_MYPICKS_NOWIKIRADIO_DESC
	DE	Der Austausch von Dateien geschieht über das WikiRadio Plugin, welches derzeit nicht installiert ist. Dieses Plugin kann <a href="http://wiki.slimdevices.com/plugin/attachments/WikiRadio/WikiRadio.pm"><u>hier</u></a> heruntergeladen werden.
	EN	File sharing is supported via the WikiRadio plugin which is not currently installed.  This plugin may be obtained <a href="http://wiki.slimdevices.com/plugin/attachments/WikiRadio/WikiRadio.pm"><u>here</u></a>.

PLUGIN_MYPICKS_WIKIRADIO_TITLECLASH
	DE	Es besteht bereits eine gleichnamige Wiedergabeliste auf WikiRadio. Bitte ändern Sie den Namen. Falls sie die existierende Datei ersetzen wollen, so besuchen Sie <a href="http://wiki.slimdevices.com/index.cgi?action=attachments&page_name=WikiRadio" target="WikiRadio"><u>das WikiRadio</u></a>, um vor dem Übertragen die bestehende Datei zu löschen.
	EN	Wiki Radio already contains a station list with this title.  Please change the title of this station list.  If you want to replace the current WikiRadio station visit <a href="http://wiki.slimdevices.com/index.cgi?action=attachments&page_name=WikiRadio" target="WikiRadio"><u>here</u></a> to delete the current file before uploading.

PLUGIN_MYPICKS_NOMYPICKS
	DE	Bitten bearbeiten sie erst ihre Internet-Radiostationen
	EN	Please use My Picks OPML Editor to create list first

';
}

1;
